//
//  $Id$
//
// Copyright (c) 2003 Daniel J. Gregor, Jr.
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
// 
//    - Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//    - Redistributions in binary form must reproduce the above
//      copyright notice, this list of conditions and the following
//      disclaimer in the documentation and/or other materials provided
//      with the distribution.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//

package com.gregor.mailping;

import java.io.*;
import java.net.*;
import java.util.*;

public class SMTPServerSocket implements Runnable {
	protected Socket s;
	protected PrintStream log;

	private static final int STATE_HELO = 1;
	private static final int STATE_MAIL = 2;
	private static final int STATE_RCPT = 3;
	private static final int STATE_DATA = 4;

	public SMTPServerSocket (Socket s) {
		this.s = s;
		this.log = System.out;
	}

	public SMTPServerSocket (Socket s, PrintStream log) {
		this.s = s;
		this.log = log;
	}

	public void run() {
		try {
			SMTPServerLoop();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void SMTPServerLoop() throws IOException {
		BufferedReader in = new BufferedReader(
		    new InputStreamReader(s.getInputStream()));
		PrintWriter out = new PrintWriter(
		    new BufferedWriter(new OutputStreamWriter(
		    s.getOutputStream())));

		// XXX is this the best way to get addresses?
		String my_address = s.getLocalAddress().getHostName();
		String their_address = s.getInetAddress().getHostName();

		String t;
                String T = null;
		String sender = null;
		Vector recipients = new Vector();
		StringBuffer message = new StringBuffer();

		int state = 1;

		out.print("220 " + my_address + " SMTPServerSocket\r\n");
		out.flush();

		while (true) {
			t = in.readLine();

			if (state != STATE_DATA) {
				T = parseCommand(t);
				if (T.equals("HELO")) {
					out.print("250 Hello " +
				            their_address + "\r\n");
					out.flush();
					continue;
				} else if (T.equals("RSET")) {
					out.print("250 Ok\r\n");
					out.flush();
					state = STATE_HELO;
					continue;
				} else if (T.equals("QUIT")) {
					out.print("221 Bye\r\n");
					out.flush();
					out.close();
					return;
				}
			}

			switch (state) {
				case STATE_HELO: {
					if (T.equals("MAIL")) {
						sender = getAddress(t);
						recipients.clear();
						message = new StringBuffer();
						out.print("250 Ok\r\n");
						out.flush();
						state = STATE_MAIL;
					} else if (T.equals("RCPT")) {
						out.print("503 need MAIL command\r\n");
						out.flush();
					} else if (T.equals("DATA")) {
						out.print("503 no valid recipients specified\r\n");
						out.flush();
					} else {
						out.print("501 Invalid command\r\n");
						out.flush();
					}
					break;
				}
				case STATE_MAIL: {
					if (T.equals("MAIL")) {
						out.print("503 nested mail commands\r\n");
						out.flush();
					} else if (T.equals("RCPT")) {
						recipients.add(
						    getAddress(t));
						out.print("250 Ok\r\n");
						out.flush();
						state = STATE_RCPT;
					} else if (T.equals("DATA")) {
						out.print("503 no valid recipients specified\r\n");
						out.flush();
					} else {
						out.print("501 Invalid command\r\n");
						out.flush();
					}
					break;
				}
				case STATE_RCPT: {
					if (T.equals("MAIL")) {
						out.print("503 nested mail commands\r\n");
						out.flush();
					} else if (T.equals("RCPT")) {
						recipients.add(
						    getAddress(t));
						out.print("250 Ok\r\n");
						out.flush();
					} else if (T.equals("DATA")) {
						out.print("354 End data with <CR><LF>.<CR><LF>\r\n");
						out.flush();
						state = STATE_DATA;
					} else {
						out.print("501 Invalid command\r\n");
						out.flush();
					}
					break;
				}
				case STATE_DATA: {
					if (t.equals(".")) {
						try {
							// XXX should I log IP addresses here?
							parseMessage(sender, (String[]) recipients.toArray(new String[0]), message.toString(), my_address, their_address);
						} catch (Exception e) {
							System.err.println("parseMessage complained: " + e.toString());
							e.printStackTrace();
						}
						out.print("250 Ok\r\n");
						out.flush();
						state = STATE_HELO;
					} else {
						message.append(t + "\r\n");
					}
					break;
				}
			}
		}
	}

	public String parseCommand(String line) {
		String command;
		int i;

		if ((i = line.indexOf(" ")) == -1) {
			command = line;
		} else {
			command = line.substring(0, i);
		}

		return command.toUpperCase();
	}

	public String getAddress(String line) {
		String address;
		int i;

		if ((i = line.indexOf(":")) == -1) {
			return null;
		}

		return line.substring(i + 1).trim();
	}

	public void parseMessage(String sender, String[] recipients,
	    String message, String my_address, String their_address)
	    throws Exception
	{
		long timestamp_l, now;
		int i, j;

		if ((i = message.indexOf("Timestamp ")) == -1) {
			throw new Exception("did not find timestamp in message from \"" + sender);
		}
		i += "Timestamp ".length();

		if ((j = message.substring(i).indexOf(" ")) == -1 &&
		    (j = message.substring(i).indexOf("\r\n")) == -1) {
			throw new Exception("did not find whitespace after timestamp in message from \"" + sender);
		}

		timestamp_l = Long.parseLong(message.substring(i, i + j));
		now = System.currentTimeMillis();

		log.println("message from " + sender + " to " +
		    TextUtils.join(", ", recipients) +
 		    " sent " + timestamp_l + " received " +
		    now + " delay " + (now - timestamp_l) +
		    " delivered to " + my_address + " from " + their_address);
	}
}
